﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using SixLabors.ImageSharp;
using SixLabors.ImageSharp.Formats.Png;

namespace PV178_WinForms_Demo
{
    public static class ImageUtilities
    {
        /// <summary>
        /// Saves the <paramref name="image"/> to the given <paramref name="path"/> on a local filesystem.
        /// The correct image format is deduced from the file extension in <paramref name="path"/>.
        /// </summary>
        /// <param name="image">The image to save</param>
        /// <param name="path"></param>
        public static void SaveImage(Image image, string path) => image.Save(path);

        /// <summary>
        /// Saves the <paramref name="image"/> to the given <paramref name="path"/> on a local filesystem.
        /// The correct image format is deduced from the file extension in <paramref name="path"/>.
        /// </summary>
        /// <param name="image">The image to save</param>
        /// <param name="path"></param>
        public static Task SaveImageAsync(Image image, string path) => image.SaveAsync(path);

        /// <summary>
        /// Extracts image data from the given <paramref name="uri"/>. Returns null if unsuccessful.
        /// </summary>
        /// <param name="uri"></param>
        /// <returns></returns>
        public static Image ExtractImage(Uri uri) => GetImageFromUri(uri, false).Result;

        /// <summary>
        /// Extracts image data from the given <paramref name="uri"/>. Returns null if unsuccessful.
        /// </summary>
        /// <param name="uri"></param>
        /// <returns></returns>
        public static Task<Image> ExtractImageAsync(Uri uri) => GetImageFromUri(uri, true);

        /// <summary>
        /// Extracts image data from the given <paramref name="path"/>. Returns null if unsuccessful.
        /// </summary>
        /// <param name="uri"></param>
        /// <returns></returns>
        public static Image ExtractImage(string path) => Image.Load(path);

        /// <summary>
        /// Extracts image data from the given <paramref name="path"/>. Returns null if unsuccessful.
        /// </summary>
        /// <param name="uri"></param>
        /// <returns></returns>
        public static Task<Image> ExtractImageAsync(string path) => Image.LoadAsync(path);

        /// <summary>
        /// Extracts image data from the given <paramref name="stream"/>. Returns null if unsuccessful.
        /// </summary>
        /// <param name="uri"></param>
        /// <returns></returns>
        public static Image ExtractImage(Stream stream) => Image.Load(stream);

        /// <summary>
        /// Extracts image data from the given <paramref name="stream"/>. Returns null if unsuccessful.
        /// </summary>
        /// <param name="uri"></param>
        /// <returns></returns>
        public static Task<Image> ExtractImageAsync(Stream stream) => Image.LoadAsync(stream);

        /// <summary>
        /// Extracts image data from the given <paramref name="dataObject"/>. Returns null if unsuccessful.
        /// </summary>
        /// <param name="uri"></param>
        /// <returns></returns>
        public static Image ExtractImage(IDataObject dataObject) =>
            GetImageFromFileDrop(dataObject, false).Result ??
            GetImageFromBitmap(dataObject, false).Result ??
            GetImageFromUri(dataObject, false).Result;

        /// <summary>
        /// Extracts image data from the given <paramref name="dataObject"/>. Returns null if unsuccessful.
        /// </summary>
        /// <param name="uri"></param>
        /// <returns></returns>
        public static async Task<Image> ExtractImageAsync(IDataObject dataObject) =>
            await GetImageFromFileDrop(dataObject, true).ConfigureAwait(false) ??
            await GetImageFromBitmap(dataObject, true).ConfigureAwait(false) ??
            await GetImageFromUri(dataObject, true).ConfigureAwait(false);

        private static Task<Image> GetImageFromFileDrop(IDataObject dataObject, bool executeAsync)
        {
            if (!dataObject.GetDataPresent(DataFormats.FileDrop))
            {
                return Task.FromResult<Image>(null);
            }

            var filePath = (dataObject.GetData(DataFormats.FileDrop) as string[])?.FirstOrDefault();
            if (filePath is null)
                return Task.FromResult<Image>(null);

            if (executeAsync)
            {
                return Image.LoadAsync(filePath);
            }

            return Task.FromResult(Image.Load(filePath));
        }

        private static Task<Image> GetImageFromBitmap(IDataObject dataObject, bool executeAsync)
        {
            if (!dataObject.GetDataPresent(DataFormats.Bitmap))
            {
                return Task.FromResult<Image>(null);
            }

            if (dataObject.GetData(DataFormats.Bitmap) is not System.Drawing.Bitmap bitmap)
                return Task.FromResult<Image>(null);

            if (executeAsync)
            {
                return bitmap.ToImageSharpImageAsync();
            }

            return Task.FromResult(bitmap.ToImageSharpImage());
        }

        private static Task<Image> GetImageFromUri(IDataObject dataObject, bool executeAsync)
        {
            if (!Uri.TryCreate(dataObject.GetData(DataFormats.Text) as string, UriKind.Absolute, out var uri))
                return Task.FromResult<Image>(null);
            return GetImageFromUri(uri, executeAsync);
        }

        private static async Task<Image> GetImageFromUri(Uri uri, bool executeAsync)
        {
            using var client = new HttpClient();

            if (executeAsync)
            {
                using var response = await client.GetAsync(uri).ConfigureAwait(false);
                using var stream = await response.Content.ReadAsStreamAsync().ConfigureAwait(false);
                return await Image.LoadAsync(stream).ConfigureAwait(false);
            }
            else
            {
                // Intentional indent in order to be able to reuse variable names
                using var response = client.Send(new HttpRequestMessage(HttpMethod.Get, uri));
                using var stream = response.Content.ReadAsStream();
                return Image.Load(stream);
            }
        }

        /// <summary>
        /// Checks whether the given web <paramref name="url"/> is pointing to an image resource.
        /// </summary>
        /// <param name="url"></param>
        /// <returns></returns>
        public static bool IsUrlImage(string url) => IsUrlImage(url, false).Result;

        /// <summary>
        /// Checks whether the given web <paramref name="url"/> is pointing to an image resource.
        /// </summary>
        /// <param name="url"></param>
        /// <returns></returns>
        public static Task<bool> IsUrlImageAsync(string url) => IsUrlImage(url, true);

        private static async Task<bool> IsUrlImage(string url, bool executeAsync)
        {
            if (!Uri.TryCreate(url, UriKind.Absolute, out var uri))
                return false;

            using var client = new HttpClient();

            if (executeAsync)
            {
                using var response = await client.GetAsync(uri).ConfigureAwait(false);
                using var stream = await response.Content.ReadAsStreamAsync().ConfigureAwait(false);
                return (await Image.IdentifyAsync(stream).ConfigureAwait(false)) is not null;
            }
            else
            {
                using var response = client.Send(new HttpRequestMessage(HttpMethod.Get, uri));
                using var stream = response.Content.ReadAsStream();
                return Image.Identify(stream) is not null;
            }
        }

        /// <summary>
        /// Checks whether the given <paramref name="dataObject"/> contains an image.
        /// </summary>
        /// <param name="dataObject"></param>
        /// <returns></returns>
        public static bool IsImage(IDataObject dataObject) => IsImage(dataObject, false).Result;

        /// <summary>
        /// Checks whether the given <paramref name="dataObject"/> contains an image.
        /// </summary>
        /// <param name="dataObject"></param>
        /// <returns></returns>
        public static Task<bool> IsImageAsync(IDataObject dataObject) => IsImage(dataObject, true);

        /// <summary>
        /// Checks whether the given <paramref name="dataObject"/> contains an image.
        /// </summary>
        /// <param name="dataObject"></param>
        /// <returns></returns>
        public static async Task<bool> IsImage(IDataObject dataObject, bool executeAsync)
        {
            if (dataObject.GetDataPresent(DataFormats.Bitmap))
            {
                return true;
            }
            else if (dataObject.GetDataPresent(DataFormats.FileDrop))
            {
                return (await Image.IdentifyAsync((dataObject.GetData(DataFormats.FileDrop) as string[])?.FirstOrDefault()).ConfigureAwait(false)) is not null;
            }
            else if (dataObject.GetDataPresent(DataFormats.Text))
            {
                return await IsUrlImageAsync(dataObject.GetData(DataFormats.Text) as string).ConfigureAwait(false);
            }
            return false;
        }

        /// <summary>
        /// Converts an ImageSharp <paramref name="image"/> to an equvalent <see cref="System.Drawing.Image"/>.
        /// </summary>
        /// <param name="image"></param>
        /// <returns></returns>
        public static System.Drawing.Image ToSystemDrawingImage(this Image image)
        {
            if (image is null)
                return null;

            using MemoryStream ms = new();
            image.Save(ms, Configuration.Default.ImageFormatsManager.FindEncoder(PngFormat.Instance));
            ms.Position = 0;
            return System.Drawing.Image.FromStream(ms);
        }

        /// <summary>
        /// Converts an System.Drawing <paramref name="image"/> to an equvalent <see cref="Image"/>.
        /// </summary>
        /// <param name="image"></param>
        /// <returns></returns>
        public static Image ToImageSharpImage(this System.Drawing.Image image)
        {
            if (image is null)
                return null;

            using MemoryStream ms = new();
            image.Save(ms, System.Drawing.Imaging.ImageFormat.Png);
            ms.Position = 0;
            return Image.Load(ms);
        }

        /// <summary>
        /// Converts an System.Drawing <paramref name="image"/> to an equvalent <see cref="Image"/>.
        /// </summary>
        /// <param name="image"></param>
        /// <returns></returns>
        public static Task<Image> ToImageSharpImageAsync(this System.Drawing.Image image)
        {
            if (image is null)
                return Task.FromResult<Image>(null);

            using MemoryStream ms = new();
            image.Save(ms, System.Drawing.Imaging.ImageFormat.Png);
            ms.Position = 0;
            return Image.LoadAsync(ms);
        }
    }
}