﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using SixLabors.ImageSharp;

namespace PV178_WinForms_Demo
{
    public partial class PhotoFilterForm : Form
    {
        private Image _originalImage;

        private Image OriginalImage
        {
            get => _originalImage;
            set
            {
                pbOriginal.Image?.Dispose();
                _originalImage?.Dispose();

                _originalImage = value;
                pbOriginal.Image = value?.ToSystemDrawingImage();

                EditedImage = Filter.Apply(value, Strength);
            }
        }

        private Image _editedImage;

        private Image EditedImage
        {
            get => _editedImage;
            set
            {
                pbPreview.Image?.Dispose();
                _editedImage?.Dispose();

                _editedImage = value;
                pbPreview.Image = value.ToSystemDrawingImage();
            }
        }

        private IImageFilter Filter => cbPhotoFilter.SelectedItem as IImageFilter;

        private IImageFilter[] AvailableFilters { get; } = typeof(Program).Assembly.GetTypes()
            .Where(x =>
                !x.IsAbstract &&
                typeof(IImageFilter).IsAssignableFrom(x) &&
                x.GetConstructor(Type.EmptyTypes) is not null)
            .Select(x => Activator.CreateInstance(x) as IImageFilter)
            .OrderBy(x => x.Name)
            .ToArray();

        private float _strength;

        private float Strength
        {
            get => _strength;
            set
            {
                _strength = value;
                EditedImage = Filter.Apply(OriginalImage, value);
            }
        }

        public PhotoFilterForm()
        {
            InitializeComponent();
            pbOriginal.AllowDrop = true;
            lblStengthValue.DataBindings.Add(nameof(lblStengthValue.Text), tbBlend, nameof(TrackBar.Value), true);
            cbPhotoFilter.DataSource = AvailableFilters;

            RegisterMenuItemEvents();
        }

        private void RegisterMenuItemEvents()
        {
            openFromFileToolStripMenuItem.Click += OpenFromFileToolStripMenuItem_Click;
            openFromClipboardToolStripMenuItem.Click += OpenFromClipboardToolStripMenuItem_Click;
            importFromWebToolstripMenuItem.Click += ImportFromWebToolstripMenuItem_Click;
            saveToolStripMenuItem.Click += SaveToolStripMenuItem_Click;
            exitToolStripMenuItem.Click += ExitToolStripMenuItem_Click;
        }

        private void ExitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        private void SaveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (EditedImage is null)
                return;

            SaveFileDialog dialog = new()
            {
                AddExtension = true,
                DefaultExt = ".png",
                RestoreDirectory = true,
                Filter = "PNG Image (*.png)|*.png|JPEG Image (*.jpg;*jpeg)|(*.jpg;*jpeg)|BMP Image (*.bmp)|*.bmp"
            };

            if (dialog.ShowDialog(this) != DialogResult.OK)
                return;

            EditedImage.Save(dialog.FileName);
        }

        private async void ImportFromWebToolstripMenuItem_Click(object sender, EventArgs e)
        {
            var dialog = new ImportFromUrlForm();
            if (dialog.ShowDialog(this) != DialogResult.OK)
                return;

            if (!await ImageUtilities.IsUrlImageAsync(dialog.Url))
            {
                MessageBox.Show("The URL does not point to an image.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                return;
            }

            OriginalImage = await ImageUtilities.ExtractImageAsync(new Uri(dialog.Url));
        }

        private async void OpenFromFileToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFileDialog dialog = new()
            {
                Multiselect = false,
                RestoreDirectory = true,
                Filter = "PNG Image (*.png)|*.png|JPEG Image (*.jpg;*jpeg)|(*.jpg;*jpeg)|BMP Image (*.bmp)|*.bmp"
            };

            if (dialog.ShowDialog(this) != DialogResult.OK)
                return;

            OriginalImage = await ImageUtilities.ExtractImageAsync(dialog.FileName);
        }

        private async void OpenFromClipboardToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OriginalImage = await ImageUtilities.ExtractImageAsync(Clipboard.GetDataObject());
        }

        private async void pbOriginal_DragEnter(object sender, DragEventArgs e)
        {
            if (await ImageUtilities.IsImageAsync(e.Data))
            {
                e.Effect = DragDropEffects.Copy;
            }
        }

        private async void pbOriginal_DragDrop(object sender, DragEventArgs e)
        {
            OriginalImage = await ImageUtilities.ExtractImageAsync(e.Data);
        }

        private void tbBlend_ValueChanged(object sender, EventArgs e)
        {
            Strength = tbBlend.Value / (float)(tbBlend.Maximum - tbBlend.Minimum);
        }

        private void cbPhotoFilter_SelectedValueChanged(object sender, EventArgs e)
        {
            if (OriginalImage is not null)
                EditedImage = Filter.Apply(OriginalImage, Strength);
        }
    }
}